/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Scanner;
import java.util.UUID;

import org.json.JSONException;

public class FSStorage extends AbstractTreeStorage {
    private static final String ROOT_NODE_FILENAME = "root.json";
    private Map<UUID, File> nodeToFile = new HashMap<UUID, File>();
    private String path;

    public FSStorage(String path) {
        if (!path.endsWith(File.separator)) {
            path = path + File.separator;
        }
        this.path = path;
    }

    private UUID getUUId(File file) {
        try {
            return NodeMapper.getUUId(readFile(file));
        } catch (JSONException e) {
            e.printStackTrace();
            return null;
        }
    }

    public File getFile(UUID uuid) {
        return nodeToFile.get(uuid);
    }

    private String readFile(File file) {
        try {
            StringBuilder jsonsb = new StringBuilder();
            Scanner s = new Scanner(new FileInputStream(file), "UTF-8");
            while (s.hasNextLine()) {
                jsonsb.append(s.nextLine().trim());
            }
            s.close();
            return jsonsb.toString();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }

    @Override
    public UUID getRootNode() {
        File file = new File(path + ROOT_NODE_FILENAME);
        if (file.exists()) {
            UUID uuid = getUUId(file);
            nodeToFile.put(uuid, file);
            return uuid;
        }
        return null;
    }

    public File getResourceFolder(UUID uuid) {
        String id = getNodeId(uuid);
        File resFolder = new File(getFile(uuid).getParent() + File.separator
                + id + "_resources");
        return resFolder;
    }

    @Override
    public String getNodeId(UUID uuid) {
        if (uuid.equals(getRootNode())) {
            return "/";
        }
        if (getFile(uuid) == null) {
            return null; // node not found
        }
        String jext = "json";
        String fileName = getFile(uuid).getName();
        return fileName.substring(0, fileName.length() - (jext.length() + 1));
    }

    @Override
    public UUID getParent(UUID uuid) {
        if (getFile(uuid) == null) {
            return null; // node not found
        }
        return getParent(getFile(uuid));
    }

    public UUID getParent(File file) {
        if (file.getAbsoluteFile().equals(path)) {
            return null;
        }

        String folderName = file.getParentFile().getName();
        String parentId = folderName.substring(0, folderName.length());

        File parentFolder = file.getParentFile().getParentFile();
        File parentResource = new File(parentFolder.getAbsoluteFile()
                + File.separator + parentId + ".json");

        if (!parentResource.exists()) {
            return null;
        }
        return getUUId(parentResource);
    }

    @Override
    public Collection<UUID> getChildren(UUID uuid) {
        Collection<UUID> result = new HashSet<UUID>();
        if (getFile(uuid) == null) {
            return result; // node not found
        }

        String nodeId = getNodeId(uuid);
        if (nodeId.equals("/")) {
            nodeId = "root";
        }
        File nodeFolder = getFile(uuid).getParentFile();
        File childrenFolder = new File(nodeFolder.getAbsoluteFile()
                + File.separator + nodeId);

        if (!childrenFolder.exists()) {
            return result;
        }

        for (File res : childrenFolder.listFiles()) {
            if (res.isFile() && res.getName().endsWith(".json")) {
                UUID childUUID = getUUId(res);
                result.add(getUUId(res));
                nodeToFile.put(childUUID, res);
            }
        }
        return result;
    }

    @Override
    public void createNode(UUID parentUUId, UUID uuid, String id,
            Collection<Attribute> attrs) {
        try {
            File file;
            if (parentUUId == null) {
                file = new File(path + ROOT_NODE_FILENAME);
            } else {
                String parentId = getNodeId(parentUUId);
                if (parentId.equals("/")) {
                    parentId = "root";
                }
                File childFolder = new File(getFile(parentUUId).getParentFile()
                        .getAbsoluteFile() + File.separator + parentId);

                if (!childFolder.exists()) {
                    childFolder.mkdirs();
                }
                file = new File(childFolder.getAbsolutePath() + File.separator
                        + id + ".json");
            }

            if (!file.getParentFile().exists()) {
                file.getParentFile().mkdirs();
            }
            if (!file.exists()) {
                file.createNewFile();
            }
            FileOutputStream out = new FileOutputStream(file);
            Writer w = new OutputStreamWriter(out, "UTF-8");
            String jsonString = NodeMapper.toJSON(uuid, attrs).toString(4);
            w.append(jsonString);
            w.flush();
            w.close();

            // update model
            nodeToFile.put(uuid, file);
            getDeltaTarget().created(uuid, true);

        } catch (UnsupportedEncodingException e) {
            e.printStackTrace();
        } catch (JSONException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void remove(UUID uuid) {
        // TODO: implement me
        // String id = getNodeId(uuid);
        // File res = getFile(uuid);
        // try {
        // new File(res.getParentFile()+id);
        // res.delete(false, null);
        // } catch (CoreException e) {
        // e.printStackTrace();
        // }
    }

    @Override
    public void changeNodeId(UUID node, String id) {
        // TODO Auto-generated method stub

    }

    @Override
    public Collection<Attribute> getAttributes(UUID uuid) {
        try {
            return NodeMapper.getAttributes(readFile(getFile(uuid)));
        } catch (JSONException e) {
            e.printStackTrace();
            return null;
        }
    }

    @Override
    public void saveAttributes(UUID node, Collection<Attribute> attributes) {
        try {
            OutputStreamWriter out = new OutputStreamWriter(
                    new FileOutputStream(getFile(node)), "UTF-8");
            String jsonStrring = NodeMapper.toJSON(node, attributes)
                    .toString(4);
            out.append(jsonStrring);
            out.close();
        } catch (UnsupportedEncodingException e) {
            e.printStackTrace();
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (JSONException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void moveNodeTo(UUID node, UUID parent) {
        // TODO Auto-generated method stub

    }

    @Override
    public boolean canLoad(UUID uuid) {
        File f = nodeToFile.get(uuid);
        if (f != null) {
            return f.exists();
        }
        // TODO Check json
        return false;
    }

    @Override
    public String getNodeResourcePath(UUID uuId) {
        return getResourceFolder(uuId).getAbsolutePath();
    }

    @Override
    public InputStream getResourceContents(UUID uuid, String resName) {
        // TODO: stream reuse check
        try {
            return new FileInputStream(new File(getNodeResourcePath(uuid)
                    + File.separator + resName));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
        return null;
    }

    @Override
    public void setResourceContents(UUID uuid, String resName, InputStream in, Object data) {
        try {
            FileOutputStream out = new FileOutputStream(new File(
                    getNodeResourcePath(uuid) + File.separator + resName));
            byte[] buffer = new byte[1024];
            int len;
            while ((len = in.read(buffer)) != -1) {
                out.write(buffer, 0, len);
            }
            out.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    @Override
    public boolean isResourceExists(UUID uuid, String resName) {
        return new File(
                getNodeResourcePath(uuid) + File.separator + resName).exists();
    }

    @Override
    public void deleteResource(UUID uuid, String resName) {
        new File(getNodeResourcePath(uuid) + File.separator + resName).delete();
    }

    
    private void recursiveResourceList(File folder, String path, ArrayList<String> result) {
        for(File f: folder.listFiles()) {
            // use unix path separator
            String curPath = path+"/"+f.getName();
            if(f.isDirectory()) {
                recursiveResourceList(f, curPath, result);
            } else {
                result.add(curPath);
            }
        }
    }

    @Override
    public String[] listResources(UUID uuid) {
        ArrayList<String> result = new ArrayList<String>();
        recursiveResourceList(getResourceFolder(uuid), "", result);
        return result.toArray(new String[0]);
    }

    @Override
    public String getName() {
        return new File(path).getName();
    }

    @Override
    public void createStorage(String name) {
        new File(path).getParentFile().mkdirs();
    }

    @Override
    public void deleteStorage() {
        // TODO Implement deleteStorage in FSStorage
        throw new UnsupportedOperationException("not implemented");
    }
}
