/*
 * Copyright 2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.core;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.UUID;
import java.util.concurrent.CopyOnWriteArrayList;

import com.unitesk.requality.core.transaction.ITransactionListener;
import com.unitesk.requality.core.transaction.TransactionStorage;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.NodeType;
import com.unitesk.requality.nodetypes.Requirement;

public class TreeDB {
    public static final String ORDER_ATTRIBUTE = "_order";
    public static final String INDEX_ATTRIBUTE = "_index";
    private List<ITreeChangeListener> listeners = new CopyOnWriteArrayList<ITreeChangeListener>();
    private TreeNode root;
    private Map<UUID, TreeNode> cache = new HashMap<UUID, TreeNode>();
    private ITreeStorage storage;
    private IUUIDGenerator uuidGenerator;
    private Map<String, NodeType> defaultTypes = new HashMap<String, NodeType>();
    {
        NodeType dbroot = new NodeType("DBRoot", TreeDB.this);
        dbroot.setClass(com.unitesk.requality.core.TreeNode.class);
        defaultTypes.put("DBRoot", dbroot);
        NodeType nodetype = new NodeType("NodeType", TreeDB.this);
        nodetype.setClass(com.unitesk.requality.nodetypes.NodeType.class);
        defaultTypes.put("NodeType", nodetype);
        NodeType node = new NodeType("Node", TreeDB.this);
        node.setClass(com.unitesk.requality.core.TreeNode.class);
        defaultTypes.put("Node", node);
    };

    private TreeDB() {
    }

    // //////////////////////////////////////////////////////////////////////////
    //
    // API methods
    //
    // //////////////////////////////////////////////////////////////////////////
    public TreeNode createChildNode(UUID pUUId, String id, String type) {
        UUID uuid = uuidGenerator.nextUUId();
        return createChildNode(uuid, pUUId, id, type);
    }

    public TreeNode createChildNode(UUID uuid, UUID pUUId, String id,
            String type) {
        if (getNode(pUUId).findChild(id) != null) {
            throw new IllegalArgumentException("Child with id: "
                    + " exists in " + getNode(pUUId).getQualifiedId());
        }
        storage.createNode(pUUId, uuid, type, id);
        return getNode(uuid);
    }

    public void removeNode(TreeNode node) {
        // Remove children first
        while (node.sizeChildren() > 0) {
            removeNode(node.getChild(0));
        }
        UUID uuid = node.getUUId();
        storage.remove(uuid);
        node.setDisposed();
    }

    public void moveNode(TreeNode node, TreeNode newParent) {
        storage.moveNodeTo(node.getUUId(), newParent.getUUId());
    }

    void setId(TreeNode node, String id) {
        if (!node.getId().equals(id)) {
            storage.changeNodeId(node.getUUId(), id);
        }
    }

    /**
     * @return root node of TreeDB
     */
    public TreeNode getRootNode() {
        if (root == null) {
            UUID rootUUID;
            if((rootUUID = storage.getRootNode())==null) { // root doesn't exists  
                return null;
            }
            root = getNode(rootUUID);
        }
        return root;
    }

    public synchronized TreeNode getNode(UUID uuid) {
    	TreeNode node = cache.get(uuid); 
        if (node != null) {
            return node;        	
        }
        if (storage.canLoad(uuid)) {
            return createFromStorage(uuid, storage.getNodeType(uuid));
        } else {
            throw new RuntimeException("node not found: " + uuid);
        }
    }

    public TreeNode getNode(String qId) {
        if (qId.equals("/")) {
            return getRootNode();
        }
        StringTokenizer st = new StringTokenizer(qId, "/");
        TreeNode cur = getRootNode();
        while (st.hasMoreTokens()) {
            String nodeId = st.nextToken();
            cur = cur.findChild(nodeId);
            if (cur == null) {
                return null;
            }
        }
        return cur;
    }

    public void saveAttributes(UUID uuid, Collection<Attribute> attributes) {
        storage.saveAttributes(uuid, attributes);
    }

    public void setResourceContent(TreeNode node, String name, InputStream in, Object data) {
        getStorage().setResourceContents(node.getUUId(), name, in, data);
    }

    public void deleteResource(TreeNode node, String name) {
        getStorage().deleteResource(node.getUUId(), name);
    }

    public boolean isResourceExists(TreeNode node, String name) {
        return getStorage().isResourceExists(node.getUUId(), name);
    }

    public String getName() {
        return storage.getName();
    }

    // //////////////////////////////////////////////////////////////////////////
    //
    // Transaction API methods
    //
    // //////////////////////////////////////////////////////////////////////////
    private ArrayList<ITransactionListener> transactionListeners = new ArrayList<ITransactionListener>();

    public void startTransaction(String name) {
        if (storage instanceof TransactionStorage) {
            throw new IllegalStateException("transaction \"" + getName()
                    + "\"is started");
        }
        TransactionStorage tStorage = new TransactionStorage(storage,
                transactionListeners);
        tStorage.setName(name);
        tStorage.setDeltaTarget(storageDeltaListener);
        storage.setDeltaTarget(nullDeltaListener);
        storage = tStorage;
    }

    public TransactionStorage commit() {
        if (!(storage instanceof TransactionStorage)) {
            throw new IllegalStateException("transaction not started");
        }
        TransactionStorage tStorage = (TransactionStorage) storage;
        tStorage.apply();
        tStorage.setDeltaTarget(nullDeltaListener);
        storage = tStorage.getStorage();
        storage.setDeltaTarget(storageDeltaListener);
        return tStorage;
    }

    public void rollback() {
        if (!(storage instanceof TransactionStorage)) {
            throw new IllegalStateException("transaction not started");
        }
        TransactionStorage tStorage = (TransactionStorage) storage;
        tStorage.setDeltaTarget(nullDeltaListener);
        storage = tStorage.getStorage();
        storage.setDeltaTarget(storageDeltaListener);
        tStorage.rollback();
        for (ITransactionListener tl : transactionListeners) {
            tl.undo(tStorage);
        }
    }

    public void addTransactionListener(ITransactionListener l) {
        transactionListeners.add(l);
    }

    public void removeTransactionListener(ITransactionListener l) {
        transactionListeners.remove(l);
    }

    // ////////////////////////////////////////////////////////////////////////
    //
    // Types operations
    //
    // ////////////////////////////////////////////////////////////////////////
    public void addNodeType(String name, Class<? extends TreeNode> clazz) {
        NodeType nt = new NodeType(name, TreeDB.this);
        nt.setClass(clazz);
        defaultTypes.put(name, nt);
    }

    public void addNodeType(Class<? extends TreeNode> clazz) {
        addNodeType(clazz.getSimpleName(), clazz);
    }

    public void addNodeTypes(Map<String, Class<? extends TreeNode>> nts) {
        for (String name : nts.keySet()) {
            addNodeType(name, nts.get(name));
        }
    }

    public Map<String, NodeType> getTypes() {
        Map<String, NodeType> result = new HashMap<String, NodeType>(
                defaultTypes);
        TreeNode typesNode = getNode("/Types");
        if (typesNode != null && typesNode.sizeChildren() > 0) {
            for (int i = 0; i < typesNode.sizeChildren(); i++) {
                NodeType node = (NodeType) typesNode.getChild(i);
                result.put(node.getId(), node);
            }
        }
        return result;
    }

    public void createType(NodeType newType) {
        TreeNode types = getNode("/Types");
        if (types == null) {
            types = getRootNode().createChild("Types", NodeType.TYPE_NAME);
        }
        TreeNode newTypeNode = types.createChild(newType.getName());
        newTypeNode.saveAttributes();
    }

    public NodeType getNodeType(String name) {
        for (NodeType type : getTypes().values()) {
            if (type.getId().equals(name)) {
                return type;
            }
        }
        return null;
    }

    public boolean removeType(String typeName) {
        TreeNode types = getNode("/Types");
        if (types == null) {
            return false;
        }
        TreeNode n = types.findChild(typeName);
        n.delete();
        return true;
    }

    // ////////////////////////////////////////////////////////////////////////
    //
    // TreeDB factories
    //
    // ////////////////////////////////////////////////////////////////////////

    /**
     * Creates db structures and root node
     * 
     * @param name
     * @param storage
     * @param uuidGenerator
     * @return
     */
    public static TreeDB create(String name, ITreeStorage storage,
            IUUIDGenerator uuidGenerator) {
        TreeDB newDB = new TreeDB();
        newDB.uuidGenerator = uuidGenerator;
        storage.createStorage(name);
        storage.setDeltaTarget(newDB.storageDeltaListener);
        newDB.storage = storage;
        // create root node
        UUID uuid = uuidGenerator.nextUUId();
        storage.createNode(null, uuid, "DBRoot", "/");
        return newDB;
    }

    /**
     * Creates db structures and root node
     * 
     * @param name
     * @param storage
     * @return treeDB object
     */
    public static TreeDB create(String name, ITreeStorage storage) {
        return create(name, storage, new DefaultUUIDGenerator());
    }

    public static TreeDB load(ITreeStorage storage, IUUIDGenerator uuidGenerator) {
        TreeDB newDB = new TreeDB();
        newDB.uuidGenerator = uuidGenerator;
        // FIXME: for test we can use mock target, so we left old one if it is
        // not null. Its look like bad design.
        if (storage.getDeltaTarget() == null) {
            storage.setDeltaTarget(newDB.storageDeltaListener);
        }
        newDB.storage = storage;
        return newDB;
    }

    public static TreeDB load(ITreeStorage storage) {
        return load(storage, new DefaultUUIDGenerator());
    }

    // ////////////////////////////////////////////////////////////////////////
    public boolean addListener(ITreeChangeListener listener) {
        return listeners.add(listener);
    }

    public boolean removeListener(ITreeChangeListener listener) {
        return listeners.remove(listener);
    }

    public List<ITreeChangeListener> getListeners() {
        return Collections.unmodifiableList(listeners);
    }

    private IStorageDeltaListener nullDeltaListener = new IStorageDeltaListener() {
        @Override
        public void moved(UUID uuid, UUID oldParent, UUID newParent,
                boolean internal) {
        }

        @Override
        public void deleted(UUID uuid, boolean internal) {
        }

        @Override
        public void created(UUID uuid, boolean internal) {
        }

        @Override
        public void changed(UUID uuid, boolean internal) {
        }
    };

    private void notifyExternalChange() {
        for (ITransactionListener tl : transactionListeners) {
            tl._dispose();
        }
    }

    private boolean compareNodes(TreeNode node, NodeDesc nodeDesc) {
        if (!node.getUUId().equals(nodeDesc.getUUId()))
            return false;
        if (node.getParentUUId() == null && nodeDesc.getParentUUId() != null || node.getParentUUId() != null && nodeDesc.getParentUUId() == null)
            return false;
        if (node.getParentUUId() != null && nodeDesc.getParentUUId() != null && !node.getParentUUId().equals(nodeDesc.getParentUUId()))
            return false;
        if (!node.getId().equals(nodeDesc.getId()))
            return false;
        
        Map<String, Attribute> oldMap = node.getAttributes();
        Map<String, Attribute> newMap = new HashMap<String, Attribute>(nodeDesc.getAttributes());
        for (String attr_key : oldMap.keySet()) {
            Attribute oldAttr = oldMap.get(attr_key);
            Attribute newAttr = null;
            if (newMap.containsKey(attr_key))
                newAttr = newMap.remove(attr_key);
            else
                return false;
            if (oldAttr == null && newAttr != null || oldAttr != null && newAttr == null)
                return false;
            if (oldAttr != null && newAttr != null && !oldAttr.equals(newAttr))
                return false;
        }
        if (!newMap.isEmpty())
            return false;
        
        List<UUID> oldList = node.getChildrenUUIds();
        List<UUID> newList = new ArrayList<UUID>(nodeDesc.getChildrenUUId());
        for (UUID uuid : oldList) {
            if (newList.contains(uuid))
                newList.remove(uuid);
            else
                return false;
        }
        if (!newList.isEmpty())
            return false;
        return true;
    }

    private IStorageDeltaListener storageDeltaListener = new IStorageDeltaListener() {
        @Override
        public void created(UUID uuid, boolean forceInternal) {
            TreeNode curNode = cache.get(uuid);
            NodeDesc newNodeDesc = loadNodeDescription(uuid);
            if (curNode != null) {
                if (!compareNodes(curNode, newNodeDesc)) {
                    throw new IllegalStateException("node with uuid: "
                            + curNode.getUUId() + "(" + curNode.getQualifiedId()
                            + ") exists");
                } else {
                    return;
                }
            }

            curNode = createModelNode(
                    newNodeDesc.getUUId(), 
                    newNodeDesc.getParentUUId(), 
                    newNodeDesc.getId(), 
                    (String)newNodeDesc.getAttributes().get(TreeNode.ATTR_TYPE).getValue(),
                    newNodeDesc.getAttributes().values(),
                    newNodeDesc.getChildrenUUId());
            if (curNode.getParent() != null 
                    && !curNode.getParent().children.contains(uuid)) {
                curNode.getParent().children.add(uuid); 
            }
            cache(uuid, curNode);
            if (curNode.getId().equals("/")) {
                assert curNode.parent == null;
                root = curNode;
            }
            if (!forceInternal) {
                notifyExternalChange();
            }
            for (ITreeChangeListener l : listeners) {
                l.created(curNode, forceInternal);
            }
        }

        private NodeDesc loadNodeDescription(UUID uuid) {
            String typeName = storage.getNodeType(uuid);
            UUID parentUUId;
            String id;
            if (typeName.equals("DBRoot")) {
                id = "/";
                parentUUId = null;
            } else {
                id = storage.getNodeId(uuid);
                parentUUId = storage.getParent(uuid);
            }
            Map<String, Attribute> attributes = new TreeMap<String, Attribute>();
            Collection<Attribute> attrs = storage.getAttributes(uuid);
            for (Attribute a : attrs) {
                attributes.put(a.getKey(), a);
            }
            List<UUID> childrenUUId = new ArrayList<UUID>(
                    storage.getChildren(uuid));
            return new NodeDesc(uuid, parentUUId, id, attributes, childrenUUId);
        }

        @Override
        public void deleted(UUID uuid, boolean internal) {
            TreeNode node = cache.get(uuid);
            NodeDesc nodeDesc = NodeDesc.fromNode(node);
            node.getParent().children.remove(uuid);
            cache.remove(uuid);
            node.setDisposed();
            if (!internal) {
                notifyExternalChange();
            }
            for (ITreeChangeListener l : listeners) {
                l.deleted(nodeDesc, internal);
            }
        }

        @Override
        public void changed(UUID uuid, boolean internal) {
            // Sometimes (e.g, import of existing project)
        	// changed event happens when 
        	// there is no node in the cache yet.
        	// So, use getNode(uuid) instead cache.get(uuid).
        	TreeNode curNode = getNode(uuid);
        	String newId = storage.getNodeId(uuid);
        	if (curNode instanceof Requirement && !newId.equals(curNode.id)) {
        	    ((Requirement)curNode).updateTreeNodes((Requirement)curNode);
        	}
            curNode.id = newId;
            loadNodeAttributes(uuid, curNode);
            if (!internal) {
                notifyExternalChange();
            }
            fireNodeChange(curNode, internal);
        }

        @Override
        public void moved(UUID nodeUUID, UUID oldParentUUId, UUID newParentUUId,
                boolean internal) {
            TreeNode node = cache.get(nodeUUID);
            TreeNode newParent = cache.get(newParentUUId);
            TreeNode oldParent = cache.get(oldParentUUId);

            node.parent = newParentUUId;
            if(oldParent!=null)
                oldParent.children.remove(nodeUUID);
            newParent.children.add(nodeUUID);

            if (!internal) {
                notifyExternalChange();
            }
            for (ITreeChangeListener l : listeners) {
                l.moved(node, oldParent == null ? node.getParent() : oldParent,
                        internal);
            }
        }
    };

    private void loadNodeAttributes(UUID uuid, TreeNode node) {
        Collection<Attribute> newAttributes = storage.getAttributes(uuid);
        applyNodeAttributes(node, newAttributes);
    }

    @SuppressWarnings("unchecked")
    private void applyNodeAttributes(TreeNode node, Collection<Attribute> newAttributes) {
        //for new and old Locations of Requirement reset labels
        if (node instanceof Requirement) {
            Requirement req = (Requirement) node;
            String oldName = req.getName();
            String newName = null;
            //get old Locations list
            List<String> oldLocs = new ArrayList<String>(req.getLocationQIds());
            //construct new Locations list
            List<String> newLocs = null;
            for (Attribute attr : newAttributes) {
                if (attr.getKey().equals(Requirement.ATTR_LOCATIONS)) {
                    Object val = attr.getValue();
                    if (val instanceof List) {
                        newLocs = new ArrayList<String>((List<String>) val);
                    }
                } else
                if (attr.getKey().equals("_name")) {
                    newName = (String) attr.getValue();
                }
            }
            //compare lists
            boolean fNewLoc = false;
            if (newLocs != null) {
                for (String newLoc : newLocs) {
                    if (!oldLocs.remove(newLoc)) {
                        fNewLoc = true;
                    }
                }
            }
            //and call affected Locations
            if (fNewLoc || !oldLocs.isEmpty()) {
                //remaining old locations (removed)
                for (String oldLoc : oldLocs) {
                    TreeNode locNode = getNode(oldLoc);
                    if (locNode instanceof Location) {
                        ((Location)locNode).resetLabel();
                    }
                }
                //all other locations
                if (newLocs != null) {
                    for (String newLoc : newLocs) {
                        TreeNode locNode = getNode(newLoc);
                        if (locNode instanceof Location) {
                            ((Location)locNode).resetLabel();
                        }
                    }
                }
            }
            //check, if name was changed
            if (oldName != null || newName != null) {
                if (oldName == null && newName != null || oldName != null && newName == null || !oldName.equals(newName)) {
                    req.updateTreeNodes(req);
                }
            }
        }
        updateNodeAttributes(node, newAttributes);
    }

    private void updateNodeAttributes(TreeNode node, Collection<Attribute> newAttributes) {
        HashSet<String> removeKeys = new HashSet<String>(
                node.attributes.keySet());
        for (Attribute a : newAttributes) {
            removeKeys.remove(a.getKey());
            node.attributes.put(a.getKey(), a);
        }
        for (String key : removeKeys) {
            node.attributes.remove(key);
        }
    }

    public IUUIDGenerator getUuidGenerator() {
        return uuidGenerator;
    }

    private void cache(UUID uuid, TreeNode node) {
        cache.put(uuid, node);
    }

    public ITreeStorage getStorage() {
        return storage;
    }

    private TreeNode createModelNode(UUID uuid, UUID puuid, String id,
            String typeName, Collection<Attribute> attributes, Collection<UUID> children) {
        TreeNode result = null;
        if (defaultTypes.containsKey(typeName)) {
            NodeType nt = defaultTypes.get(typeName);
            try {
                Class<?>[] treeNodeArgs = new Class[] { TreeDB.class,
                        UUID.class };
                result = nt.getTypeClass().getConstructor(treeNodeArgs)
                        .newInstance(TreeDB.this, uuid);
                result.setType(typeName);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        if (result == null) {
            result = new TreeNode(this, uuid);
            result.setType(typeName);
        }
        result.id = id;
        result.parent = puuid;

        result.children = new ArrayList<UUID>(children);
        updateNodeAttributes(result, attributes);
        return result;
    }

    private TreeNode createFromStorage(UUID uuid, String typeName) {
        String id;
        UUID puuid;
        if (typeName.equals("DBRoot")) {
            id = "/";
            puuid = null;
        } else {
            id = storage.getNodeId(uuid);
            puuid = storage.getParent(uuid);
        }
        Collection<Attribute> attributes = storage.getAttributes(uuid);
        Collection<UUID> children = storage.getChildren(uuid);
        TreeNode result = createModelNode(uuid, puuid, id, typeName, attributes, children);
        cache(uuid, result);
        if (result.getId().equals("/")) {
            assert result.parent == null;
            root = result;
        }
        return result;
    }

    public void fireNodeChange(TreeNode curNode, boolean internal) {
        for (ITreeChangeListener l : listeners) {
            l.changed(curNode, internal);
        }
    }

}
