/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.io.File;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.regex.Pattern;

import com.unitesk.requality.tools.Index;

public class TreeNode {
    public static final String ATTR_TYPE = "_type";
    public static final String ATTR_NAME = "_name";
    public static final String ATTR_PREDICATE = "_predicate";
    public static final String ATTR_DESCRIPTION = "_description";
    public static final String ATTR_BASE_NODES = "_baseNodes";

    private UUID uuid;
    private final TreeDB db;
    protected String id;
    UUID parent;
    List<UUID> children = new ArrayList<UUID>();
    Map<String, Attribute> attributes = new LinkedHashMap<String, Attribute>();
    private boolean disposed = false;

    public TreeNode(TreeDB db, UUID uuid) {
        if (uuid == null)
            uuid = UUID.randomUUID();
        this.uuid = uuid;
        this.db = db;
        this.disposed = false;
    }

    public void setPredicate(String pred) {
        putAttribute(new Attribute(AttributeType.STRING, ATTR_PREDICATE, pred));
    }

    public String getPredicate() {
        if (getAttributeValue(ATTR_PREDICATE) == null) {
            return "";
        }
        return (String) getAttributeValue(ATTR_PREDICATE);
    }

    public void setType(String type) {
        putAttribute(new Attribute(AttributeType.STRING, ATTR_TYPE, type));
    }

    public final UUID getUUId() {
        return uuid;
    }

    public void setId(String id) {
        db.setId(this, id);
    }

    public String getId() {
        return id;
    }

    /**
     * @return String name of node type
     */
    public String getType() {
        return (String) getAttributeValue(ATTR_TYPE);
    }

    public void setParent(TreeNode parent) {
        db.moveNode(this, parent);
    }

    public UUID getParentUUId() {
        return parent;
    }

    public void removeChild(TreeNode node) {
        db.removeNode(node);
    }

    public void setDescription(String description) {
        putAttribute(new Attribute(AttributeType.STRING,
                ATTR_DESCRIPTION, description));
    }

    public String getDescription() {
        return getDescription(false);
    }

    public String getDescription(boolean process) {
        if (attributes.get(ATTR_DESCRIPTION) != null) {
            String result = (String) attributes.get(ATTR_DESCRIPTION)
                    .getValue();
            if (process) {
                result = processResourcesPath(result);
            }
            return result;
        }
        return null;
    }

    protected String processResourcesPath(String desc) {
        String resPath = getResourcesPath();
        String result = desc.replace("${node.resURL}", resPath);
        return result;
    }

    public TreeDB getTreeDB() {
        return db;
    }

    void setUUId(UUID uuid) {
        this.uuid = uuid;
    }

    public TreeNode getParent() {
        if (parent == null) {
            // no parent
            assert this.getTreeDB().getRootNode().equals(this);
            return null;
        }
        return db.getNode(parent);
    }

    public int sizeChildren() {
        return children.size();
    }

    public TreeNode getChild(int i) {
        return db.getNode(children.get(i));
    }

    public String getQualifiedId() {
        if (getParent() == null) {
            return "/";
        }
        String parentQId = getTreeDB().getNode(parent).getQualifiedId();
        if (parentQId.equals("/")) {
            return parentQId + this.getId();
        }
        return parentQId + "/" + this.getId();
    }

    public TreeNode findChild(String id) {
        for (int i = 0; i < sizeChildren(); i++) {
            if (getChild(i).getId().equals(id)) {
                return getChild(i);
            }
        }
        return null;
    }

    public List<UUID> getChildrenUUIds() {
        return Collections.unmodifiableList(children);
    }

    @Override
    public String toString() {
        return getId() + "(" + getType() + ")";
    }

    public TreeNode createChild(String id) {
        if (findChild(id) != null)
            throw new IllegalArgumentException("child '" + id
                    + "' is already exists");
        if (getType().equals("DBRoot")) {
            return db.createChildNode(uuid, id, "Node");
        }
        return db.createChildNode(uuid, id, getType());
    }

    public TreeNode createChild(String id, String type) {
        if (!this.canBeParentOf(type))
            return null;
        return db.createChildNode(uuid, id, type);
    }

    public void delete() {
        getParent().removeChild(this);
    }

    public Attribute putAttribute(Attribute a) {
        Attribute result = attributes.get(a.getKey());
        attributes.put(a.getKey(), a);
        return result;
    }

    public Attribute removeAttribute(String key) {
        return attributes.remove(key);
    }

    public Attribute getAttribute(String key) {
        return attributes.get(key);
    }

    public Object getAttributeValue(String key) {
        return getAttributeValue(key, null);
    }

    public Object getAttributeValue(String key, Object def) {
        Attribute value = attributes.get(key);
        return value != null ? value.getValue() : def;
    }

    public int getIntAttributeValue(String key, int def) {
        Attribute value = attributes.get(key);
        return value != null ? (Integer) value.getValue() : def;
    }

    public int getIntAttributeValue(String key) {
        return getIntAttributeValue(key, 0);
    }

    public float getFloatAttributeValue(String key, float def) {
        Attribute value = attributes.get(key);
        return value != null ? (Float) value.getValue() : def;
    }

    public float getFloatAttributeValue(String key) {
        return getFloatAttributeValue(key, 0);
    }

    public boolean getBooleanAttributeValue(String key, boolean def) {
        Attribute value = attributes.get(key);
        return value != null ? (Boolean) value.getValue() : def;
    }

    public boolean getBooleanAttributeValue(String key) {
        return getBooleanAttributeValue(key, false);
    }

    public String getStringAttributeValue(String key, String def) {
        Attribute value = attributes.get(key);
        return value != null ? (String) value.getValue() : def;
    }

    public String getStringAttributeValue(String key) {
        return getStringAttributeValue(key, null);
    }

    @SuppressWarnings("unchecked")
    public List<String> getListAttributeValue(String key) {
        return (List<String>) getAttributeValue(key, null);
    }

    public Set<String> getAttributeKeys() {
        return Collections.unmodifiableSet(attributes.keySet());
    }

    public Map<String, Attribute> getAttributes() {
        return Collections.unmodifiableMap(attributes);
    }

    public void saveAttributes() {
        db.saveAttributes(uuid, this.attributes.values());
    }

    @SuppressWarnings("unchecked")
    public void addBaseNode(String nodeQId) {
        List<String> baseNodes = (List<String>) getAttributeValue(ATTR_BASE_NODES);
        if (baseNodes == null) {
            baseNodes = new ArrayList<String>();
        }
        baseNodes.add(nodeQId);
        putAttribute(new Attribute(AttributeType.LIST, ATTR_BASE_NODES,
                baseNodes));
    }

    @SuppressWarnings("unchecked")
    public boolean removeBaseNode(String nodeQId) {
        List<String> baseNodes = (List<String>) getAttributeValue(ATTR_BASE_NODES);
        if (baseNodes == null || !baseNodes.remove(nodeQId)) {
            return false;
        }
        putAttribute(new Attribute(AttributeType.LIST, ATTR_BASE_NODES,
                baseNodes));
        return true;
    }

    @SuppressWarnings("unchecked")
    public List<String> getBaseNodes() {
        if (getAttributeValue(ATTR_BASE_NODES) == null) {
            return Collections.EMPTY_LIST;
        }
        return Collections
                .unmodifiableList((List<String>) getAttributeValue(ATTR_BASE_NODES));
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null)
            return false;
        if (super.equals(obj))
            return true;
        if (!(obj instanceof TreeNode))
            return false;
        TreeNode node = (TreeNode) obj;
        if (!node.getUUId().equals(uuid))
            return false;
        if (!getQualifiedId().equals(node.getQualifiedId()))
            return false;
        if (!getAttributeKeys().equals(node.getAttributeKeys()))
            return false;
        for (String key : getAttributeKeys()) {
            if (!getAttributeValue(key).equals(node.getAttributeValue(key)))
                return false;
        }
        return true;
    }

    public TreeNode[] getSortedChildren() {
        return getSortedChildren(null, null);
    }

    public TreeNode[] getSortedChildren(String type) {
        return getSortedChildren(null, type);
    }
    
    public TreeNode[] getSortedChildren(Comparator<TreeNode> c, String type) {
        ArrayList<TreeNode> result = new ArrayList<TreeNode>();
        for (int i = 0; i < sizeChildren(); i++) {
            TreeNode ch = getChild(i);
            if (type == null || ch.getType().equals(type)) {
                result.add(getChild(i));
            }
        }
        TreeNode[] r = (TreeNode[]) result.toArray(new TreeNode[0]);
        if (c != null) {
            Arrays.sort(r, c);
        } else {
            Arrays.sort(r, new Comparator<TreeNode>() {
                @Override
                public int compare( TreeNode o1, TreeNode o2 )
                {
                    if (o1 == null || o2 == null) 
                    {
                        // FIXME: why o1 or o2 can be null?
                        return 0;
                    }
                    int k = o1.getIndex().compareTo( o2.getIndex() );
                    if (k == 0)
                        return o1.getId().compareTo(o2.getId());
                    return k;
                }
            });
        }
        return r;
    }

    public String getResourcesPath() {
        try {
            String path = new File(db.getStorage().getNodeResourcePath(uuid)).toURI()
                    .toURL().toString();
            if (path.length() > 1 && path.endsWith("/")) {
                path = path.substring(0, path.length() - 1);
            }
            return path;
        } catch (MalformedURLException e) {
            e.printStackTrace();
        }
        return null;
    }
    private static Pattern filepattern = Pattern.compile(
            "# Match a valid Windows filename (unspecified file system).          \n" +
            "^                                # Anchor to start of string.        \n" +
            "(?!                              # Assert filename is not: CON, PRN, \n" +
            "  (?:                            # AUX, NUL, COM1, COM2, COM3, COM4, \n" +
            "    CON|PRN|AUX|NUL|             # COM5, COM6, COM7, COM8, COM9,     \n" +
            "    COM[1-9]|LPT[1-9]            # LPT1, LPT2, LPT3, LPT4, LPT5,     \n" +
            "  )                              # LPT6, LPT7, LPT8, and LPT9...     \n" +
            "  (?:\\.[^.]*)?                  # followed by optional extension    \n" +
            "  $                              # and end of string                 \n" +
            ")                                # End negative lookahead assertion. \n" +
            "([^<>:\"/\\\\|?*\\x00-\\x1F\\ \\xA0]  # First char is not a space.  \n" +
            "[^<>:\"/\\\\|?*\\x00-\\x1F]*)?     # Zero or more valid filename chars.\n" +
            "[^<>:\"/\\\\|?*\\x00-\\x1F\\ \\xA0.]  # Last char is not a space or dot.  \n" +
            "$                                # Anchor to end of string.            ", 
            Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.COMMENTS);
    public static Boolean isValidFileName(String name){
        return filepattern.matcher(name).matches() && !(name.startsWith(" ") || name.endsWith(" "));
    }
    public static Boolean isValidId(TreeNode req,TreeNode parent,String text){
        if(text.trim().length() == 0)
            return false;
        if(parent == null)
            parent = req.getParent();
        if(req != null && !parent.canBeParentOf(req))
            return false;
        for (TreeNode r : parent.getSortedChildren()) {
            if (r.getId().equals(text.trim()) && (req == null || !r.getUUId().equals(req.getUUId())))
                return false;
        }
        return isValidFileName(text);
    }
    
    public static Boolean isValidId(TreeNode req,String text){
        return isValidId(req, null, text);
    }
    
    public static Boolean isValidId(String text){
        return isValidFileName(text);
    }

    public NodeDesc toNodeDesc() {
        return toNodeDesc(true);
    }

    public NodeDesc toNodeDesc(Boolean deep) {
        return NodeDesc.fromNode(this, deep);
    }

    public InputStream getResourceContent(String name) {
        return db.getStorage().getResourceContents(uuid, name);
    }

    public void setResourceContent(String name, InputStream in) {
        db.setResourceContent(this, name, in, null);
    }

    public void setResourceContent(String name, InputStream in, Object data) {
        db.setResourceContent(this, name, in, data);
    }

    public void deleteResource(String name) {
        db.deleteResource(this, name);
    }

    public boolean isResourceExists(String name) {
        return db.isResourceExists(this, name);
    }
    
    protected Index index;

    public Index getIndex()
    {
        if( index == null )
        {
            String indexStr = (String)getAttributeValue( TreeDB.INDEX_ATTRIBUTE );
            if( indexStr == null ) indexStr = (String)getAttributeValue( TreeDB.ORDER_ATTRIBUTE ); // old behaviour 
            if( indexStr == null ) indexStr = getId();
            index = Index.create( indexStr );
        }
        return index;
    }

    public void setIndex( Index index )
    {
        Attribute indexAttr = new Attribute( AttributeType.STRING, TreeDB.INDEX_ATTRIBUTE, index.toString() );
        putAttribute( indexAttr );
        this.index = index;
    }

    public boolean isDisposed() {
        return disposed;
    }

    public void setDisposed() {
        disposed = true;
    }
    /**
     * Check if target node can become child of current TreeNode
     * @param child node to check possibility to add it as child
     * @return true if target can be a child of current node 
     */
    public boolean canBeParentOf(TreeNode child){
        return canBeParentOf(child.getType());
    }
    public boolean canBeParentOf(String child){
        List<String> list = possibleChildren();
        if(list == null)
            return true;
        if(list.size() == 0)
            return false;
        return list.contains(child);
    }
    /**
     * Check if target node can become parent of current node
     * @param parent node to check possibility to set it as parent
     * @return true if target node can be a parent of current node 
     */
    public boolean canBeChildOf(TreeNode parent){
        List<String> list = possibleParents();
        if(list == null)
            return true;
        if(list.size() == 0)
            return false;
        return list.contains(parent.getType());
    }
    protected List<String> possibleParents(){
        return null;
    }
    protected List<String> possibleChildren(){
        return null;
    }
    public List<String> getPossibleChildTypes(){
        List<String> list=possibleChildren();
        if(list == null)
            return new ArrayList<String>();
        return list;
    }
    public List<String> getPossibleParentTypes(){
        List<String> list=possibleParents();
        if(list == null)
            return new ArrayList<String>();
        return list;
    }
    public boolean isCopyable(){
        return false;
    }
    public boolean canBeCopiedTo(TreeNode parent){
        return canBeChildOf(parent);
    }
    public boolean canAddCopyOf(TreeNode child){
        return canBeParentOf(child);
    }
    public boolean createCopyIn(TreeNode target){
        return false;
    }
    public boolean canBeDeleted(){
        return true;
    }
    
    public String getUserFriendlyName() {
        TreeNode parent = (TreeNode) getParent();
        String userFriendlyName = parent.getUserFriendlyName();
        if (userFriendlyName.length() > 0) {
            return userFriendlyName + "/" + getId();
        }
        return getId();
    }

    public void updateTreeNodes(TreeNode req) {
        TreeNode[] reqs = req.getSortedChildren();
        for (TreeNode node : reqs) {
            updateTreeNodes(node);
        }
        req.getTreeDB().fireNodeChange(req, true);
    }
}
