package com.unitesk.requality.core.transaction;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;
import java.util.UUID;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.ITreeStorage;
import com.unitesk.requality.core.AbstractTreeStorage;

public class TransactionStorage extends AbstractTreeStorage {
    public class OperationData {
        private TreeOperation operation;
        private Object data;

        public OperationData(TreeOperation operation, Object data) {
            this.operation = operation;
            this.data = data;
        }

        public TreeOperation getOperation() {
            return operation;
        }

        public Object getData() {
            return data;
        }

    }
    private static final String TMP_RESOURCE_PREFIX = "requality";
    private ITreeStorage storage;
    private Map<UUID, String> newId = new HashMap<UUID, String>();
    private Map<UUID, String> oldId = new HashMap<UUID, String>();
    private Map<UUID, Collection<UUID>> newChildren = new HashMap<UUID, Collection<UUID>>();
    private Map<UUID, Collection<UUID>> oldChildren = new HashMap<UUID, Collection<UUID>>();
    private Map<UUID, Collection<Attribute>> newAttr = new HashMap<UUID, Collection<Attribute>>();
    private Map<UUID, Collection<Attribute>> oldAttr = new HashMap<UUID, Collection<Attribute>>();
    private Map<UUID, HashMap<String, String>> resourсes = new LinkedHashMap<UUID, HashMap<String, String>>();

    private LinkedList<TreeOperation> operations = new LinkedList<TreeOperation>();
    private LinkedList<OperationData> operationsData = new LinkedList<OperationData>();

    private List<ITransactionListener> tListeners;

    private String transactionName = "complex operation";
    private boolean execute = true;

    public TransactionStorage(ITreeStorage storage, List<ITransactionListener> tListeners) {
        this.storage = storage;
        this.tListeners = tListeners;
    }

    public String getName() {
        return transactionName;
    }

    public void setName(String name) {
        transactionName = name;
    }

    @Override
    public UUID getRootNode() {
        return storage.getRootNode();
    }

    @Override
    public UUID getParent(UUID uuid) {
        for (UUID pUUId : newChildren.keySet()) {
            if (newChildren.get(pUUId).contains(uuid)) {
                return pUUId;
            }
        }
        UUID oldParentUUID = getOldParent(uuid);
        if (oldParentUUID != null) { // if node was deleted
            return oldParentUUID;
        }
        return storage.getParent(uuid);
    }

    @Override
    public String getNodeId(UUID uuid) {
        if (newId.containsKey(uuid)) {
            return newId.get(uuid);
        }
        return storage.getNodeId(uuid);
    }

    @Override
    public Collection<Attribute> getAttributes(UUID uuid) {
        return currentAttrs(uuid);
    }

    @Override
    public Collection<UUID> getChildren(UUID uuid) {
        return currentChildren(uuid);
    }

    @Override
    public boolean canLoad(UUID uuid) {
        if(getParent(uuid) == null) {
            return false;
        }
        return currentChildren(getParent(uuid)).contains(uuid);
    }

    @Override
    public String getNodeResourcePath(UUID uuId) {
        // TODO: implement me
        return storage.getNodeResourcePath(uuId);
    }

    @Override
    public String[] listResources(UUID uuid) {
        return currentResources(uuid).keySet().toArray(new String[0]);
    }

    public ITreeStorage getStorage() {
        return storage;
    }

    private UUID getOldParent(UUID uuid) {
        for (UUID pUUId : oldChildren.keySet()) {
            if (oldChildren.get(pUUId).contains(uuid)) {
                return pUUId;
            }
        }
        return null;
    }

    @Override
    public void createNode(UUID parentUUId, UUID uuid, String id,
            Collection<Attribute> attrs) {
        newId.put(uuid, id);
        oldId.put(uuid, null);
        currentChildren(parentUUId).add(uuid);

        newAttr.put(uuid, attrs);
        oldAttr.put(uuid, new TreeSet<Attribute>());

        CreateNode createOp = new CreateNode(parentUUId, uuid, id, attrs);
        operations.addLast(createOp);

        getDeltaTarget().created(uuid, true);
    }

    @Override
    public void remove(UUID uuid) {
        UUID pUUID = getParent(uuid);
        String id = getNodeId(uuid);
        Collection<Attribute> attrs = getAttributes(uuid);

        newId.remove(uuid);
        oldId.put(uuid, id);
        currentChildren(pUUID).remove(uuid);

        // TODO: transaction remove resources
        Map<String, String> resources = currentResources(uuid);
        for(String key: resources.keySet()) {
            deleteResource(pUUID, key);
        }
        DeleteNode deleteOperation = new DeleteNode(pUUID, uuid, id, attrs);
        operations.addLast(deleteOperation);

        getDeltaTarget().deleted(uuid, true);
    }

    @Override
    public void changeNodeId(UUID uuid, String newNodeId) {
        String oldNodeId = getNodeId(uuid);
        oldId.put(uuid, oldNodeId);
        newId.put(uuid, newNodeId);
        RenameNode moveOperation = new RenameNode(uuid, oldNodeId, newNodeId);
        operations.addLast(moveOperation);
        
        getDeltaTarget().changed(uuid, true);
    }

    @Override
    public void saveAttributes(UUID uuid, Collection<Attribute> newAttributes) {
        Collection<Attribute> oldAttributes = currentAttrs(uuid);
        newAttr.put(uuid, newAttributes);
        SaveAttributes saveOperation = new SaveAttributes(uuid, oldAttributes,
                newAttributes);
        operations.addLast(saveOperation);
        getDeltaTarget().changed(uuid, true);
    }

    @Override
    public void moveNodeTo(UUID uuid, UUID newParent) {
        UUID oldParent = getParent(uuid);
        currentChildren(oldParent).remove(uuid);
        currentChildren(newParent).add(uuid);
        MoveNode moveOperation = new MoveNode(uuid, oldParent, newParent);
        operations.addLast(moveOperation);
        getDeltaTarget().moved(uuid, oldParent, newParent, true);
    }

    private Collection<UUID> currentChildren(UUID uuid) {
        if(uuid==null) {
            throw new IllegalArgumentException("uuid is null");
        }
        if (!newChildren.containsKey(uuid)) {
            TreeSet<UUID> initalData = new TreeSet<UUID>();
            if (storage.getChildren(uuid) != null) {
                initalData.addAll(storage.getChildren(uuid));
            }
            newChildren.put(uuid, new TreeSet<UUID>(initalData));
            oldChildren.put(uuid, new TreeSet<UUID>(initalData));
        }
        return newChildren.get(uuid);
    }

    private Collection<Attribute> currentAttrs(UUID uuid) {
        if (!newAttr.containsKey(uuid)) {
            HashSet<Attribute> initalData = new HashSet<Attribute>();
            if (storage.getAttributes(uuid) != null) {
                initalData.addAll(storage.getAttributes(uuid));
            }
            newAttr.put(uuid, new HashSet<Attribute>(initalData));
            oldAttr.put(uuid, new HashSet<Attribute>(initalData));
        }
        return newAttr.get(uuid);
    }

    private Map<String, String> currentResources(UUID uuid) {
        if (!resourсes.containsKey(uuid)) {
            resourсes.put(uuid, new HashMap<String, String>());
        }
        return resourсes.get(uuid);
    }

    public void apply() {
        apply(storage);
    }
    public void apply(ITreeStorage aStorage) {
        operationsData.clear();

        for (TreeOperation op : operations) {
            op.apply(aStorage);
            operationsData.add(new OperationData(op, op.getData()));
        }
        for(ITransactionListener tl: tListeners) {
            if(execute) {
                tl.execute(this);
            } else {
                tl.redo(this);
            }
        }
        execute = false; // next time apply means redo
    }

    public void rollback() {
        rollback(storage);
    }

    public void rollback(ITreeStorage aStorage) {
        operationsData.clear();

        LinkedList<TreeOperation> reverceOperations = new LinkedList<TreeOperation>(
                operations);
        Collections.reverse(reverceOperations);
        for (TreeOperation op : reverceOperations) {
            op.rollback(aStorage);
            operationsData.add(new OperationData(op, op.getData()));
        }
        for(ITransactionListener tl: tListeners) {
            tl.undo(this);
        }
    }

    public boolean isResourceInTransaction(UUID uuid, String resName) {
        Map<String, String> nodeResources = resourсes.get(uuid);
        if (nodeResources == null) {
            return false;
        }
        if(nodeResources.containsKey(resName)) {
            return true;
        }
        return false;
    }

    // //////////////////////
    // Resources
    @Override
    public boolean isResourceExists(UUID uuid, String resName) {
        if (isResourceInTransaction(uuid, resName)) {
            if (resourсes.get(uuid).containsKey(resName) 
                    && resourсes.get(uuid).get(resName) == null) {
                return false;
            }
            return true;
        }
        if (storage.isResourceExists(uuid, resName)) {
            if (!resourсes.containsValue(uuid)) { // we doun't change node in
                                                  // transaction
                return true;
            }

            if (!resourсes.get(uuid).containsKey(resName)) { // we doun't change
                                                             // resource in
                                                             // transaction
                return true;
            }
            if (resourсes.get(uuid).get(resName) != null) {
                return true;
            }
        }
        return false;
    }

    @Override
    public InputStream getResourceContents(UUID uuid, String resName) {
        if (isResourceInTransaction(uuid, resName)) {
            try {
                if(resourсes.get(uuid).get(resName)!=null) {
                    return new FileInputStream(resourсes.get(uuid).get(resName));
                }
                return null;
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            }
        }
        return storage.getResourceContents(uuid, resName);
    }

    @Override
    public void setResourceContents(UUID uuid, String resName, InputStream in, Object data) {
        try {
            Map<String, String> nodeResources = currentResources(uuid);
          
            String oldContentFile = null;
            
            if (isResourceExists(uuid, resName)) {
                File tmpOldFile = File.createTempFile(TMP_RESOURCE_PREFIX,
                        "old");
                oldContentFile = tmpOldFile.getAbsolutePath();
                InputStream oldIn = getResourceContents(uuid, resName);

                OutputStream out = new FileOutputStream(tmpOldFile);
                ChangeResource.copy(oldIn, out );
                out.close();
                oldIn.close();
            }

            File tmpNewFile = File.createTempFile(TMP_RESOURCE_PREFIX, "new");
            String newContentFile = tmpNewFile.getAbsolutePath();
            OutputStream out = new FileOutputStream(tmpNewFile);
            ChangeResource.copy(in, out);
            nodeResources.put(resName, newContentFile);

            ChangeResource operation = new ChangeResource(uuid, resName,
                    oldContentFile, newContentFile, data);
            operations.addLast(operation);

            getDeltaTarget().changed(uuid, true);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void deleteResource(UUID uuid, String resName) {
        try {
            File tmpOldFile = File.createTempFile(TMP_RESOURCE_PREFIX, "new");
            OutputStream out = new FileOutputStream(tmpOldFile);
            String oldContentFile = tmpOldFile.getAbsolutePath();
            InputStream in = getResourceContents(uuid, resName);
            ChangeResource.copy(in , out);

            Map<String, String> nodeResources = currentResources(uuid);
            nodeResources.put(resName, null);

            ChangeResource operation = new ChangeResource(uuid, resName,
                    oldContentFile, null, null);
            operations.addLast(operation);

            getDeltaTarget().changed(uuid, true);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void dispose() {
        for(TreeOperation op: operations) {
            op.dispose();
        }
    }

    @Override
    public void createStorage(String name) {
        // TODO Implement me!
        throw new UnsupportedOperationException("not implemented");
    }

    @Override
    public void deleteStorage() {
        // TODO Implement me!
        throw new UnsupportedOperationException("not implemented");
    }

    public LinkedList<OperationData> getLastOperationsData() {
        return operationsData;
    }
}
