/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.documents;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.Stack;
import java.util.UUID;

import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.events.XMLEvent;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;

public class HTMLStructureDocumentProcessor extends DocumentProcessor
{
    @Override
    protected void process( String docId, Reader in, Writer out, TreeNode rootNode ) throws IOException
    {        
        try
        {
            XMLEventReader reader = XMLInputFactory.newInstance().createXMLEventReader( in );
            XMLEventWriter writer = XMLOutputFactory.newInstance().createXMLEventWriter( out );
            XMLEventFactory eventFactory = XMLEventFactory.newInstance();
            
            int headLevel = -1;
            int chIndex = 0;
    
            Stack<Requirement> rootNodes = new Stack<Requirement>();
            Stack<Integer> ids = new Stack<Integer>();
            rootNodes.push( (Requirement)rootNode );
            ids.push(0);
            Requirement req = null;
            Location location = null;
            String locationId = null;
            
            String name = null;
            int order = 0;
            
            while( reader.hasNext() )
            {
                XMLEvent event = (XMLEvent)reader.next();
                if( event.isStartElement() ) 
                {
                    String tagName = event.asStartElement().getName().getLocalPart();
                    int hl = headLevel( tagName );
                    if( hl > 0 ) 
                    {
                        headLevel = hl;
                        rootNodes.setSize( headLevel );
                        ids.setSize( headLevel + 1 );
                        Requirement parent = null;
                        Integer id = null;
                        for( int i = rootNodes.size() - 1; i >= 0; i-- )
                        {
                            parent = rootNodes.get(i);
                            if( parent != null )
                            {
                                id = ids.get(i+1);
                                id = ( id != null ? id : 0 ) + 1;
                                ids.set( i+1, id );
                                break;
                            }
                        }
                        req = (Requirement) parent.createChild( "" + id );
                        req.putAttribute( new Attribute( AttributeType.INT, "_order", order++ ) );
                        rootNodes.push( req );
                        Document doc = (Document) req.getTreeDB().getNode(docId);
                        location = (Location) doc.createChild(UUID.randomUUID().toString(), Location.TYPE_NAME);
                        locationId = location.getId().toString();
                        req.addLocation( location.getQualifiedId() );
                        name = "";
                        chIndex = 0;
                    } else
                    if( hl == 0 )
                    {
                        headLevel = hl;
                        int level = rootNodes.size();
                        Requirement parent = rootNodes.get(level-1);
                        if( ids.size() < level+1 ) ids.setSize( level+1 );
                        Integer id  = ids.get(level);
                        id = ( id != null ? id : 0 ) + 1;
                        ids.set( level, id );
                        req = (Requirement) parent.createChild( "" + id );
                        req.putAttribute( new Attribute( AttributeType.INT, "_order", order++ ) );
                        Document doc = (Document) req.getTreeDB().getNode(docId);
                        location = (Location) doc.createChild(UUID.randomUUID().toString(), Location.TYPE_NAME);
                        locationId = location.getId().toString();
                        req.addLocation( location.getQualifiedId() );
                        name = "";
                        chIndex = 0;
                    }
                    
                    writer.add(event);
                } else
                if( event.isEndElement() ) 
                {
                    String tagName = event.asEndElement().getName().getLocalPart();
                    int hl = headLevel( tagName );
                    if( hl > 0 ) 
                    {
                        headLevel = -1;
                        name = name.replaceAll( "\\s+", " " ).trim();
                        req.setName(name);
                        req.saveAttributes();
                        location.setDescription(name);
                        location.setHidden( true );
                        location.saveAttributes();
                        location = null;
                        req = null;
                        name = null;
                    } else 
                    if( hl == 0 )
                    {
                        headLevel = -1;
                        req.setName("");
                        req.saveAttributes();
                        location.setDescription(name);
                        location.saveAttributes();
                        location = null;
                        req = null;
                        name = null;
                    }
                    writer.add(event);
                } else
                if( event.isCharacters() )
                {
                    String data = event.asCharacters().getData();
                    if( headLevel >= 0 )
                    {
                        if( chIndex == 0 )
                        {
                            if( data.trim().length() > 0 ) chIndex++;
                        } else {
                            chIndex++;
                        }
                        if( chIndex > 0 )
                        {
                            name += data;
                            
                            writer.add( eventFactory.createStartElement("", "", "span" ) );
                            writer.add( eventFactory.createAttribute( "class", "requality_text id_" + locationId ) );
    
                            if( chIndex == 1 )
                            {
                                writer.add( eventFactory.createStartElement("", "", "a" ) );
                                writer.add( eventFactory.createAttribute( "name", locationId ) );
                                writer.add( eventFactory.createAttribute( "id", "id_" + locationId ) );
                                writer.add( eventFactory.createAttribute( "class", "requality_id" ) );
                                writer.add( eventFactory.createEndElement("", "", "a" ) );
                            }
                            
                            writer.add(event);
                            
                            writer.add( eventFactory.createEndElement("", "", "span" ) );
                        } else {
                            writer.add(event);
                        }
                        chIndex++;
                    } else {
                        writer.add(event);
                    }
                } else {
                    writer.add(event);
                }
            }
        } 
        catch( Throwable e)
        {
            throw new IOException( e );
        }
    }

    private int headLevel( String tagName )
    {
        if( "p".equalsIgnoreCase(tagName) ) return 0;
        if( "ol".equalsIgnoreCase(tagName) ) return 0;
        if( "ul".equalsIgnoreCase(tagName) ) return 0;
        if( tagName.length() != 2 ) return -1;
        char h = tagName.charAt(0);
        if( h != 'h' && h != 'H' ) return -1;
        int headLevel = tagName.charAt(1) - '0';
        if( headLevel < 1 || headLevel > 7 ) return -1;
        return headLevel;
    }
}
