/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.properties;

import java.util.UUID;

import org.eclipse.core.resources.IProject;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.views.properties.tabbed.AbstractPropertySection;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;

import com.unitesk.requality.core.ITreeChangeListener;
import com.unitesk.requality.core.NodeDesc;
import com.unitesk.requality.core.TreeChangeListener;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.core.ITreeStateListener;
import com.unitesk.requality.eclipse.core.TreesTracker;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;

public abstract class TreeNodePropertiesTab extends AbstractPropertySection {
    protected TreeNode node;
    boolean isDisposed = true;

    ITreeChangeListener treeChangeListener = new TreeChangeListener() {
        @Override
        public void moved(TreeNode newNode, TreeNode oldParent) {
            setTreeNodeSafe(newNode);
        }

        @Override
        public void deleted(NodeDesc nodeDesc) {
            if (nodeDesc.getUUId().equals(uuid)) {
                node = null;
                uuid = null;
                if(page.getControl().isDisposed()) {
                    return;
                }
                if (page.getCurrentTab() != null) {
                    ISelection selection = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getSelectionService().getSelection();
                    page.selectionChanged(getPart(), selection);
                }
            }
        }

        @Override
        public void changed(TreeNode newNode) {
            setTreeNodeSafe(newNode);
        }

        @Override
        public void created(TreeNode newNode) {
            setTreeNodeSafe(newNode);
        }
    };

    ITreeStateListener treeStateListener = new ITreeStateListener() {
        @Override
        public void open(TreeDB db) {
            db.addListener(treeChangeListener);
            setTreeNodeSafe(node);
        }

        @Override
        public void close(TreeDB db) {
            db.removeListener(treeChangeListener);
            setTreeNodeSafe(node);
        }

        @Override
        public void delete(IProject proj) {
            // TODO Auto-generated method stub

        }
    };
    private UUID uuid;
    private TabbedPropertySheetPage page;

    @Override
    public void createControls(Composite parent,
            TabbedPropertySheetPage aTabbedPropertySheetPage) {
        page = aTabbedPropertySheetPage;
        super.createControls(parent, aTabbedPropertySheetPage);
        isDisposed=false;
    }

    public TreeNodePropertiesTab() {
        super();
        for (TreeDB db : TreesTracker.getTracker().getTrees()) {
            db.addListener(treeChangeListener);
        }
        TreesTracker.getTracker().addListener(treeStateListener);
    }

    @Override
    public void dispose() {
        isDisposed = true;
        node = null;
        uuid = null;
        for (TreeDB db : TreesTracker.getTracker().getTrees()) {
            db.removeListener(treeChangeListener);
        }
        TreesTracker.getTracker().removeListener(treeStateListener);
        super.dispose();
    }

    private void setTreeNodeSafe(final TreeNode newNode) {
        if (newNode == null || node == null || node.isDisposed() || !node.getQualifiedId().equals(newNode.getQualifiedId())) {
            return;
        }
        Display d = Display.getCurrent();
        if (d == null) {
            d = Display.getDefault();
        }
        d.asyncExec(new Runnable() {
            @Override
            public void run() {
                if (!isDisposed) {
                    setTreeNode(newNode);
                }
            }
        });
    }

    abstract protected void setTreeNode(TreeNode node);

    @Override
    public void setInput(IWorkbenchPart part, ISelection selection) {
        super.setInput(part, selection);
        node = TreeNodePropertiesTab.toTreeNode(selection);
        if (node != null) {
            uuid = node.getUUId();
            setTreeNodeSafe(node);
        }
    }

    protected Requirement toRequirement(TreeNode node) {
        if (node instanceof Requirement) {
            return (Requirement) node;
        }

        if (node instanceof Location) {
            Location location = (Location) node;
            Requirement[] reqs = location.getRequirements();
            if (reqs.length > 0) {
                return reqs[0];
            }
        }
        return null;
    }

    public static TreeNode toTreeNode(ISelection sel) {
        TreeNode node = (TreeNode) ((IStructuredSelection) sel)
                .getFirstElement();
        if (node instanceof TreeNode) {
            return (TreeNode) node;
        }
        return null;
    }
}
