package com.unitesk.requality.merge;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.ranges.DocumentRange;
import org.w3c.dom.ranges.Range;

import com.unitesk.requality.merge.Chunk.InvalidLocationChunkException;
import com.unitesk.requality.merge.Tree.NodeFilter;

public class Location {

	private final String id;
	private List<Chunk> chunks;
	private Node commonParent;
	private Status status;

	public enum Status {
		INITIAL, TRACED, MISSING, MOVED, CHANGED
	};

	@SuppressWarnings("serial")
	public class TraceFailedException extends Exception {

		private String expectedText;
		private String actualText;

		public TraceFailedException() {

		}

		public TraceFailedException(String expected, String actual) {
			this.expectedText = expected;
			this.actualText = actual;
		}

		public String getExpectedValue() {
			return expectedText;
		}

		public String getActualValue() {
			return actualText;
		}
	}

	public Location(Chunk chunk) {
		status = Status.INITIAL;
		chunks = new ArrayList<Chunk>();
		chunks.add(chunk);
		id = chunk.getLocationId();
		Document spanOwner = chunk.getSpanNode().getDomNode()
				.getOwnerDocument();
		Document anchorOwner = chunk.getAnchorNode().getDomNode()
				.getOwnerDocument();
		if (!spanOwner.isSameNode(anchorOwner))
			commonParent = chunk.getAnchorNode().getParent();
		else
			commonParent = chunk.getSpanNode().getParent();
	}

	public String getLocationId() {
		return id;
	}

	public List<Chunk> getChunks() {
		return chunks;
	}

	public Node getCommonParent() {
		return commonParent;
	}

	public void addChunk(Chunk chunk) {
		if (chunk.getLocationId().equals(id)) {
			chunks.add(chunk);
			commonParent = findCommonParent();
		}
	}

	private Node findCommonParent() {
		Node lca = commonParent;
		for (Chunk chunk : chunks) {
			Node currParent = chunk.getAnchorNode().getParent();
			if (currParent != null && !lca.isSame(currParent)) {
				Context currParentCtx = currParent.getContext();
				if (!currParentCtx.isWithin(lca.getContext())) {
					if (!lca.getContext().isWithin(currParentCtx)) {
						while (!currParentCtx.isWithin(lca.getContext())) {
							lca = lca.getParent();
						}
					} else
						lca = currParent;
				}
			}
		}
		return lca;
	}

	public Location traceLocation(Tree t) throws TraceFailedException {
		Node parent = traceParent(t);
		if (parent == null) {
			status = Status.MISSING;
			throw new TraceFailedException();
		}
		Location newLoc = null;
		for (Chunk chunk : chunks) {
			Node anchor = traceChunk(chunk, parent, t);
			if (anchor != null) {
				try {
					Chunk newChunk = new Chunk(chunk, anchor);
					if (newLoc == null)
						newLoc = new Location(newChunk);
					else
						newLoc.addChunk(newChunk);
				} catch (InvalidLocationChunkException e) {
					status = Status.MISSING;
					throw new TraceFailedException();
				}
			} else {
				status = Status.CHANGED;
				throw new TraceFailedException(commonParent.getText(),
						parent.getText());
			}
		}
		if (!newLoc.commonParent.isEqual(this.commonParent)) {
			status = Status.MOVED;
			throw new TraceFailedException(commonParent.toString(),
					parent.toString());
		} else
			status = Status.TRACED;
		return newLoc;
	}

	private Node traceParent(Tree t) {
		NodeFilter parentFilter = new NodeFilter() {
			@Override
			public boolean accept(Node x, Node referenceNode) {
				if (x.isText())
					return false;
				if (Utils.jaccardSimilarity(x.getText(),
						referenceNode.getText()) < 0.3d)
					return false;
				return true;
			}
		};
		for (Node parent : t.nearestMatch(commonParent, parentFilter)) {
			String text = parent.getText();
			if (Utils
					.nonSymmetricSimilarity(this.getText().trim(), text.trim()) > 0.5d)
				return parent;
		}
		return null;
	}

	private Node traceChunk(Chunk chunk, Node parent, Tree t) {
		NodeFilter matchFilter = new NodeFilter() {
			@Override
			public boolean accept(Node x, Node referenceNode) {
				if (x.getContext().isExcluded())
					return false;
				if (x.getDomNode().getNodeType() != org.w3c.dom.Node.TEXT_NODE)
					return false;
				return true;
			}
		};
		List<Node> textMatches = t.nearestMatch(chunk.getSpanNode(), parent,
				matchFilter);
		for (Node textMatch : textMatches) {
			if (textMatch.getText().equals(chunk.getChunkText()))
				textMatch.getContext().excludeFromSearch(true);
			if (Utils.nonSymmetricSimilarity(chunk.getChunkText(),
					textMatch.getText()) > 0.5d)
				return textMatch;
		}
		return null;
	}

	public String toString() {
		String str = "";
		str += id;
		str += " : " + commonParent;
		str += " : " + commonParent.getText();
		return str;
	}

	public String getText() {
		String str = "";
		for (Chunk chunk : chunks)
			str += chunk.getChunkText() + " ";
		return str;
	}

	public Status getStatus() {
		return status;
	}
}
