/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.UUID;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.documents.DeepFirstXMLDOMWalker;

public class Document extends TreeNode {
    public static final String TYPE_NAME = "Document";
    private static final String ATTRIBUTE_FILE = "_file";
    private static final String TYPE_ROOT_QID = "/Documents";
    private List<UUID> locationCache;
    private org.w3c.dom.Document xmlDoc;

    public Document() {
        super(null, null);
    }

    private Element getElementChildIgnoreCase(Node node, String name) {
        for (int i = 0; i < node.getChildNodes().getLength(); i++) {
            if (node.getChildNodes().item(i).getNodeName().toLowerCase()
                    .equals(name.toLowerCase())) {
                return (Element) node.getChildNodes().item(i);
            }
        }
        return null;
    }

    public void setMetaCharset(String charset) {
        try {
            org.w3c.dom.Document dom = getDOM();
            Node html = dom.getFirstChild();
            Node head = getElementChildIgnoreCase(dom.getFirstChild(), "head");
            if (head == null) {
                head = dom.createElement("head");
                html.appendChild(head);
            }
            Element meta = getElementChildIgnoreCase(head, "meta");
            if (meta == null) {
                meta = dom.createElement("meta");
                head.appendChild(meta);
            }
            meta.setAttribute("http-equiv", "Content-Type");
            meta.setAttribute("content", "text/html; charset=" + charset);
            saveDOM(null);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public String getMetaCharset() {
        // TODO: check rfc for this meta
        String result = null;
        try {
            DeepFirstXMLDOMWalker dfw = new DeepFirstXMLDOMWalker(getDOM()
                    .getFirstChild());
            for (Node node : dfw) {
                if (node.getNodeName().toLowerCase().equals("meta")) {
                    if (node.getAttributes().getNamedItem("http-equiv") == null)
                        continue;
                    if (!node.getAttributes().getNamedItem("http-equiv")
                            .getNodeValue().toLowerCase()
                            .equals("content-type"))
                        continue;
                    if (node.getAttributes().getNamedItem("content") == null)
                        continue;
                    String content = node.getAttributes()
                            .getNamedItem("content").getNodeValue();
                    int i = content.indexOf("charset=");
                    if (i < 0)
                        return null;
                    return content.substring(i + "charset=".length());
                }
            }
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return result;
    }

    public static String getTypeRootQId() {
        return TYPE_ROOT_QID;
    }

    public Document(TreeDB db, UUID uuid) {
        super(db, uuid);
        setType(TYPE_NAME);
    }

    public void setResourceName(String name) {
        Attribute a = new Attribute(AttributeType.STRING, ATTRIBUTE_FILE, name);
        putAttribute(a);
    }

    @Override
    public void setParent(TreeNode parent) {
        String oldDocQId = getQualifiedId();
        super.setParent(parent);
        String newDocQId = parent.getQualifiedId() + "/" + id;
        List<UUID> locs = getLocations();
        for (UUID loc : locs) {
            String oldQid = oldDocQId + "/" + loc.toString();
            String newQid = newDocQId + "/" + loc.toString();
            Requirement reqs[] = Location.getRequirementsFor(getTreeDB(),
                    oldQid);
            for (Requirement req : reqs) {
                req.removeLocation(oldQid);
                req.addLocation(newQid);
                req.saveAttributes();
            }
        }
    }

    public String getResourceName() {
        return (String) getAttributeValue(ATTRIBUTE_FILE);
    }

    // private static final String LOCATION_CLASS_ID_PREFIX =
    // "requality_text id_";
    private static final String LOCATION_CLASS_ID_PREFIX = "id_";

    public String getHTMLPath() {
        return getTreeDB().getStorage().getNodeResourcePath(getUUId())
                + File.separator + getResourceName();
    }

    public List<Location> getLocationNodes() {
        List<Location> result = new ArrayList<Location>();
        for (UUID locQId : getLocations()) {
            TreeNode locNode = getTreeDB().getNode(locQId);
            if (locNode != null && locNode instanceof Location) {
                result.add((Location) locNode);
            }
        }
        return Collections.unmodifiableList(result);
    }
    
    public List<UUID> getLocations() {
        if (locationCache == null) {
            locationCache = new ArrayList<UUID>();
            try {
                DeepFirstXMLDOMWalker dfw = new DeepFirstXMLDOMWalker(loadDOM()
                        .getFirstChild());
                for (Node node : dfw) {
                    if (node.getNodeName().equals("span")) {
                        Node spanClass = node.getAttributes().getNamedItem(
                                "class");
                        if (spanClass != null) {
                            String classNames[] = spanClass.getNodeValue()
                                    .split(" ");
                            for (String className : classNames) {
                                if (className
                                        .startsWith(LOCATION_CLASS_ID_PREFIX)) {
                                    UUID uuid = UUID.fromString(className
                                            .substring(LOCATION_CLASS_ID_PREFIX
                                                    .length()));
                                    if (!locationCache.contains(uuid)) {
                                        locationCache.add(uuid);
                                    }
                                }
                            }
                        }
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        return locationCache;
    }

    public static class LocationRegion {
        public LocationRegion() {
            startNodeType = endNodeType = Node.ELEMENT_NODE;
        }

        public LocationRegion(String sX, int sT, int sOff, String eX, int eT, int eOff, String text) {
            startNodeXPath = sX;
            startNodeType = sT;
            startNodeOffset = sOff;
            endNodeXPath = eX;
            endNodeType = eT;
            endNodeOffset = eOff;
            this.text = text;
        }

        public boolean isStarted() {
            return startNodeXPath != null;
        }

        private String startNodeXPath;
        private int startNodeType;
        private int startNodeOffset;

        private String endNodeXPath;
        private int endNodeType;
        private int endNodeOffset;

        private String text = "";

        @Override
        public String toString() {
            String start = makeNodeString("start", startNodeXPath,
                    startNodeType, startNodeOffset);
            if(endNodeXPath==null) {
                endNodeXPath = startNodeXPath;
                endNodeOffset=startNodeOffset + text.length();
            }
            String end = makeNodeString("end", endNodeXPath, endNodeType,
                    endNodeOffset);
            return "{" + start + "," + end + ",\"text\":\"" + text + "\"}";
        }

        private String makeNodeString(String name, String xpath, int type,
                int offset) {
            StringBuilder sb = new StringBuilder();
            sb.append("\"");
            sb.append(name);
            sb.append("\":{\"node\":\"");
            sb.append(xpath);
            sb.append("\",\"nodeType\":");
            sb.append(type);
            sb.append(",\"offset\":");
            sb.append(offset);
            sb.append("}");
            return sb.toString();
        }

        public void text(String text) {
            if(!isText()) {
                startNodeType = endNodeType = Node.TEXT_NODE;
            }
            this.text+=text;
        }
        public boolean isText() {
            return startNodeType == Node.TEXT_NODE;
        }

        public void xpath(String xpath) {
            if(!isStarted()) {
                startNodeXPath = xpath;
            } else {
                endNodeXPath = xpath;
            }
        }

        public void offset(int offset) {
            if(!isStarted()) {
                startNodeOffset = offset;
            } else {
                endNodeOffset = offset;
            }
        }
    }

    public String removeLocation(String uuid) {
        try {
            List<LocationRegion> regions = new ArrayList<LocationRegion>();
            removeLocation(UUID.fromString(uuid), getDOM().getFirstChild(),
                    regions, 0);

            StringBuilder sb = new StringBuilder();
            sb.append("{\"action\":\"delete\",\"uuid\":\"");
            sb.append(uuid);
            sb.append("\",\"selection\":[");
            boolean first = true;
            for (LocationRegion r : regions) {
                if (!first) {
                    sb.append(",");
                    first=false;
                }
                sb.append(r.toString());
            }
            sb.append("]}");
            String result = sb.toString(); 
            saveDOM(result);

            return result;
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    private static boolean needDropNode(Node node, UUID uuid) {
        if (node.getNodeName().equals("a")) {
            if (node.getAttributes().getNamedItem("id").getNodeValue()
                    .equals("id_" + uuid)) {
                return true;
            }
        }

        if (node.getNodeName().equals("span")) {
            String className = node.getAttributes().getNamedItem("class")
                    .getNodeValue();
            if (className.startsWith("requality_link requality_link_" + uuid)) {
                return true;
            }
            if (className
                    .startsWith("requality_link requality_link_in_" + uuid)) {
                return true;
            }
        }

        return false;
    }

    public Location createLocation() {
        UUID locUUId = UUID.randomUUID();
        return (Location) getTreeDB().createChildNode(locUUId, getUUId(), locUUId.toString(), Location.TYPE_NAME);
    }


    // TODO: nested location support
    private void removeLocation(UUID uuid, Node node,
            List<LocationRegion> regions, final int parentOffset) {
        int len = 0;
        if (isLocationNode(node)) {
            if (getUUID(node).equals(uuid)) {
                LocationRegion r = currentRegion(regions);
                Element parentElement = (Element) node.getParentNode();

                // delete nodes
                for (int i = 0; i < node.getChildNodes().getLength(); i++) {
                    Node curNode = node.getChildNodes().item(i);
                    if (needDropNode(curNode, uuid)) {
                        continue;
                    }

                    if(curNode.getNodeType()==Node.TEXT_NODE) {
                        len = curNode.getTextContent().length();
                        r.text(curNode.getTextContent());
                    }

                    parentElement.insertBefore(curNode,
                            node.getNextSibling());

                }
                parentElement.removeChild(node);

                String xpath = getElementXpath(parentElement);
                if(r.isStarted()) {
                    r.offset(parentOffset+len);
                } else {
                    r.offset(parentOffset);
                }
                r.xpath(xpath);
            }
        }

        int offset=0;
        for (int i = 0; i < node.getChildNodes().getLength(); i++) {
            Node child = node.getChildNodes().item(i);
            if(child.getNodeType() == Node.TEXT_NODE) {
                int nodeLen = child.getTextContent().length();
                offset+=nodeLen;
            }
            removeLocation(uuid, child, regions, offset);
        }
    }

    private LocationRegion currentRegion(List<LocationRegion> regions) {
        if (regions.size() == 0) {
            regions.add(new LocationRegion());
        }
        return regions.get(regions.size() - 1);
    }

    public void saveDOM() {
        saveDOM(null);
    }
    public void saveDOM(String jsData) {
        ByteArrayOutputStream newXml = new ByteArrayOutputStream();
        try {
            org.w3c.dom.Document doc = getDOM();
            Writer out = new OutputStreamWriter(newXml, "UTF-8");
            Result result = new StreamResult(out);
            TransformerFactory.newInstance().newTransformer()
                    .transform(new DOMSource(doc), result);
            out.close();
            setResourceContent(getResourceName(), new ByteArrayInputStream(
                    newXml.toByteArray()), jsData);

        } catch (Exception e) {
            e.printStackTrace();
        }
        // TODO: optimize cache
        resetCache();
    }

    private boolean isLocationNode(Node node) {
        if (node.getNodeName().equals("span")) {
            Node spanClass = node.getAttributes().getNamedItem("class");
            if (spanClass == null) {
                return false;
            }
            String className = spanClass.getNodeValue();
            if (className.contains(LOCATION_CLASS_ID_PREFIX)
                    && className.contains("requality_text")) {
                return true;
            }
        }
        return false;
    }

    private UUID getUUID(Node locationNode) {
        try {
            String classNames[] = locationNode.getAttributes()
                    .getNamedItem("class").getNodeValue().split(" ");
            for (String className : classNames) {
                if (className.startsWith(LOCATION_CLASS_ID_PREFIX)) {
                    return UUID.fromString(className
                            .substring(LOCATION_CLASS_ID_PREFIX.length()));
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
        return null;
    }

    public Location getLocation(String locationId) {
        return (Location) findChild(locationId);
    }

    public org.w3c.dom.Document getDOM() throws Exception {
        if (getXmlDoc() == null) {
            setXmlDoc(loadDOM());
        }
        return getXmlDoc();
    }

    public org.w3c.dom.Document loadDOM() throws Exception {
        DocumentBuilder docBuilder = DocumentBuilderFactory.newInstance()
                .newDocumentBuilder();
        InputStream in = getResourceContent(getResourceName());
        org.w3c.dom.Document result = docBuilder.parse(in);
        in.close();
        return result;
    }

    private boolean containsLocationId(String[] classNames) {
        boolean requalityTextClass = false;
        boolean locationIdClass = false;

        for (String name : classNames) {
            if (name.equals("requality_text")) {
                requalityTextClass = true;
            }
            if (name.startsWith(LOCATION_CLASS_ID_PREFIX)) {
                locationIdClass = true;
            }
            if (requalityTextClass && locationIdClass) {
                return true;
            }
        }
        return requalityTextClass && locationIdClass;
    }

    private String getLocationIdClass(String[] classNames) {
        for (String name : classNames) {
            if (name.startsWith(LOCATION_CLASS_ID_PREFIX)) {
                return name;
            }
        }
        return null;
    }

    public String extractLocationText(String locUUID) {
        try {
            DocumentFragment fragment = getDOM().createDocumentFragment();

            DeepFirstXMLDOMWalker dfw = new DeepFirstXMLDOMWalker(getDOM()
                    .getFirstChild());
            for (Node node : dfw) {
                if (node.getNodeName().equals("span")) {
                    Node spanClass = node.getAttributes().getNamedItem("class");
                    if (spanClass != null) {
                        String[] classNames = spanClass.getNodeValue().split(
                                " ");

                        if (containsLocationId(classNames)) {
                            String locationIdClass = getLocationIdClass(classNames);
                            UUID uuid = UUID.fromString(locationIdClass
                                    .substring(LOCATION_CLASS_ID_PREFIX
                                            .length()));
                            if (uuid.toString().equals(locUUID)) {
                                for (int i = 0; i < node.getChildNodes()
                                        .getLength(); i++) {
                                    Node ch = node.getChildNodes().item(i);
                                    if (ch.getNodeName().equalsIgnoreCase("a")
                                            && ch.getAttributes().getNamedItem(
                                                    "name") != null
                                            && ch.getAttributes()
                                                    .getNamedItem("name")
                                                    .getNodeValue()
                                                    .equals(locUUID)) {
                                        continue;
                                    }
                                    fragment.appendChild(ch.cloneNode(true));
                                }
                            }
                        }
                    }
                }
            }
            ByteArrayOutputStream result = new ByteArrayOutputStream();
            TransformerFactory transFactory = TransformerFactory.newInstance();
            Transformer transformer = transFactory.newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION,
                    "yes");
            transformer.transform(new DOMSource(fragment), new StreamResult(
                    result));
            result.flush();
            return result.toString("UTF-8");

        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    public int sizeLocations() {
        return getLocations().size();
    }

    @Override
    public String toString() {
        return getId();
    }

    public void resetCache() {
        // TODO: optimize cache
        locationCache = null;
    }

    protected org.w3c.dom.Document getXmlDoc() {
        return xmlDoc;
    }

    protected void setXmlDoc(org.w3c.dom.Document xmlDoc) {
        this.xmlDoc = xmlDoc;
    }

    public String createLocationRangesDescriptor(String locId) {
        try {
            getDOM();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return "";
    }

    // /////////////////////////
    // xpath stuff

    public static String getElementXpath(Element elt) {
        String path = "";

        try {
            for (; elt != null; elt = (Element) elt.getParentNode()) {
                int idx = getElementIdx(elt);
                String xname = elt.getTagName().toString();

                if (idx >= 1)
                    xname += "[" + idx + "]";
                path = "/" + "x:" + xname + path;
            }
        } catch (Exception ee) {
        }
        return path;
    }

    public static int getElementIdx(Element elt) {
        int count = 1;
        try {
            for (Element sib = (Element) elt.getNextSibling(); sib != null; sib = (Element) sib
                    .getNextSibling()) {
                if (sib.getTagName().equals(elt.getTagName())) {
                    count++;
                }
            }
        } catch (Exception ee) {
        }
        return count;
    }

    private static List<String> possparents=new ArrayList<String>();
        private static List<String> posschildren=new ArrayList<String>();
        static{
            possparents.add(DocFolder.TYPE_NAME);
            posschildren.add(Location.TYPE_NAME);
        }
    
        @Override
        protected List<String> possibleParents() {
            return possparents;
        }
    
        public boolean isCopyable(){
            //TODO document copy
            return false;
        }
    
        @Override
        protected List<String> possibleChildren() {
            return posschildren;
        }
    
}
